/*======================================================================

    das16s.c    v0.03    Apr 26 2000     (v0.01 Feb 17 2000)

    PCM-DAS16s/16 real-time PCMCIA client driver

    for Real Time Linux kernels only, at present.
    see http://www.rtlinux.org

    by Steve Rosenbluth
    (stever8@earthlink.net, stever@la.creatureshop.henson.com)

    Copyright (C) 2000 Jim Henson's Creature Shop (Los Angeles)

    The contents of this file are subject to the Mozilla Public
    License Version 1.1 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    The developer of the initial PCMCIA code is David A. Hinds
    <dhinds@pcmcia.sourceforge.org>.  Portions created by David A. Hinds
    are Copyright (C) 1999 David A. Hinds.  All Rights Reserved.    

    Thanks to Computer Boards, Inc. for providing register documentation

    ----------------------------------------------------------------------
    Revision Notes v0.02 :
    First working revision, tested w/ das-config and readdas binaries @ 50 KHz
    in burst mode.        

    Revision Notes rev0.03 :
    * added support for digital inputs
    * Tried to fix bug where a timer restart in burst mode had out-of-order
    samples in first bank - added a channel scan reset to START fop, now
    restarts work OK ?
    * Fixed bug where we didn't OR w/ mux when writing digital output byte in
    fop_ioctl()
    * Added "truth table" in fop_ioctl() to set DIO direction based on number of
    banks of analog or digital inputs requested: If either > 1 : upper nibble is
    output (2 LSBs reserved for external MUX), and if any dig ins wanted lower
    is input. If neither > 1 : if no dig ins wanted upper and lower are output,
    else when one bank of each wanted lower is input and upper is output. 
    * This revision does not use same nibbles of DIO port as 1st rev 0.02 !
    * Added support for digital input during burst mode (high nibble only)
    * Fixed "missing conversion bug". Driver is much more reliable now. At begin
    of burst we start pacer and enable irqs, and at end we stop pacer and 
    disable irqs. Timeouts are now done by port 0x80 writes.
    * Still have 2 known bugs: an "extra conversion" bug where beginning of a 
    channel scan has errant sample. "bad sample" bug where first channel value 
    is wrong. One of these might occur 3 or 4 times in a 24 hour period of 
    continuous sampling.
    * Fixed bug where starting a burst in middle of previous burst corrupted
    data. Now we ignore the second start. This seems to solve the 
    "extra conversion" bug above, and it seems to prevent driver lockup.
    * Made digital outputs shift left by 2 when an external MUX is in use.

======================================================================*/
/*  epson dot-matrix printer condense code */

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>                                                    /* for memcpy_tofs()*/

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>
#include <pcmcia/bus_ops.h>
#include "das16s.h"
/*#include "das16s-config.h"*/
#include <rtlinux/rtl_fifo.h>
#include <rtlinux/rtl_core.h>                                                       /* for rtl_hard_enable_irq, rtl_request_irq */
#include <rtlinux/rtl_sync.h>


/*
   All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
   you do not define PCMCIA_DEBUG at all, all the debug code will be
   left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
   be present but disabled -- but it can then be enabled for specific
   modules at load time with a 'pc_debug=#' option to insmod.
*/

#define PCMCIA_DEBUG 2
#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
#define DEBUG(n, args...) if (pc_debug>(n)) printk(KERN_DEBUG args)
static char *version =
"das16s.c  v0.03  April 26, 2000  (by Steve Rosenbluth)";
/*"das16s.c  v0.01  2000/01/21 (by Steve Rosenbluth)";*/
#else
#define DEBUG(n, args...)
#endif

/* works OK! a delay. used to wait for IO regs to settle. asm/io.h trusts this as a safe port to write */
/*#define WAIT_A_WHILE() outb(0,0x80)*/

/* used to wait for IO regs to settle. asm/io.h trusts this as a safe port to write */
#define WAIT_A_WHILE(num) \
  { int wait_cnt = num; \
    if (wait_cnt > 0) \
        while(wait_cnt-- > 0) \
            outb(0,0x80); }

/* clear the fifo by reading samples out of it */
#define DAS_CLEAR_FIFO(base_addr) \
    { while (inw(base_addr+DAS_CTRL_REG) & FIFO_NOT_EMPTY_MASK) \
      inw(base_addr+DAS_CONV_REG); }

/* put an analog sample in current result table */
#define INSERT_ADC_RESULT(dev,val,num) { \
        dev->result_table[dev->result_table_tail][ADC_DATA_OFFSET+num] = val; }   /* always insert data */

/* put mux stamp in current packet of result table */
#define SET_RESULT_MUX(dev,val) { \
        dev->result_table[dev->result_table_tail][HDR_MUX_OFFSET] = val; }   /* always insert data */

/* move to next result packet in table, wrapping as necessary */
#define NEXT_ADC_RESULT_PACKET(dev)                                         /* try making table larger by incrementing tail */ \
    { \
        if (dev->result_table_tail == MAX_SAMPLE_PACKETS)                   /* can't increment */ \
            dev->result_table_tail = 0;                                     /* wrap around to beginning */ \
        else \
            dev->result_table_tail += 1;                                    /* move forward to next available slot */ \
        if (dev->result_table_head == dev->result_table_tail)               /* table is full */ \
        { \
            if (dev->result_table_head == MAX_SAMPLE_PACKETS)               /* can't increment */ \
                dev->result_table_head = 0;                                 /* wrap to begin */ \
            else \
                dev->result_table_head += 1;                                /* move forward, erasing former first slot */ \
        } \
        else                                                                /* table not full */ \
            dev->result_pack_avail++;                                       /* one more packet is available in table */ \
    }

/* for reading an analog sample in result table - returns the pointer to packet */
/* Caution: must check if any samples available before requesting pointer ! */
#define GET_ADC_RESULT_PACKET(dev) &(dev->result_table[dev->result_table_head][0])

/* for reading an analog sample in result table - returns the pointer to packet */
/* Caution: must check if any samples available before requesting pointer ! */
#define GET_CURRENT_RESULT_PACKET(dev) &(dev->result_table[dev->result_table_tail][0])

#define CLEAR_ADC_RESULT(dev) \
    { \
        if (dev->result_pack_avail)                                         /* only clear if table contains something */ \
        { \
            if (dev->result_table_head == MAX_SAMPLE_PACKETS)               /* can't increment */ \
                dev->result_table_head = 0;                                 /* wrap to begin */ \
            else \
                dev->result_table_head += 1;                                /* move forward, former first slot is already erased */ \
            dev->result_pack_avail--;                                       /* one less packet available in table now */ \
        } \
    }

/* put (packed) digital input value into current packet of result table */
#define SET_DIG_IN_RESULT(dev,val) { \
        dev->result_table[dev->result_table_tail][DIG_IN_OFFSET] = val; }        

/*#define CLEAR_RESULT_TABLE(dev) \*/


/*====================================================================*/

/* Parameters that can be set with 'insmod' */

/* Release IO ports after configuration? */
static int free_ports = 0;

/* The old way: bit map of interrupts to choose from */
/* This means pick from 15, 14, 12, 11, 10, 9, 7, 5, 4, and 3 */
static u_int irq_mask = 0xdeb8;
/* Newer, simpler way of listing specific interrupts */
static int irq_list[4] = { -1 };
static u8 init_dio = 0;                                          /* this value written to digital output ports initially */

MODULE_PARM(free_ports, "i");
MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");
MODULE_PARM(init_dio, "i");
MODULE_PARM(testing, "i");


/*====================================================================*/

/*
   The event() function is this driver's Card Services event handler.
   It will be called by Card Services when an appropriate card status
   event is received.  The config() and release() entry points are
   used to configure or release a socket, in response to card
   insertion and ejection events.  They are invoked from the dummy
   event handler. 
*/

static void das16s_config(dev_link_t *link);
static void das16s_release(u_long arg);
static int das16s_event(event_t event, int priority,
		       event_callback_args_t *args);

/*
   The attach() and detach() entry points are used to create and destroy
   "instances" of the driver, where each instance represents everything
   needed to manage one actual PCMCIA card.
*/

static dev_link_t *das16s_attach(void);
static void das16s_detach(dev_link_t *);


/***************** prototypes **********************/
int das16s_reset(void);
/*int das16s_ctl_fifo_handler(unsigned int fifo); */                          /* handler invoked in rt task when fifo accessed         */
int das16s_fop_open(struct inode *inode_p, struct file *file_p);
ssize_t das16s_fop_read(struct file *filp, char *buf, size_t count, loff_t *offset_p);
int das16s_fop_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg);
int das16s_fop_release(struct inode *inode_p, struct file *file_p);
void das16s_program_timers(unsigned int adc_divisor,unsigned int aux_adc_divisor, bool mode);
int das16s_program_adc(void);
unsigned int das16s_irq_handler(unsigned int irq, struct pt_regs *regs);


/*
   The dev_info variable is the "key" that is used to match up this
   device driver with appropriate cards, through the card configuration
   database.
*/

static dev_info_t dev_info = "das16s";

/*
   A linked list of "instances" of the das16s device.  Each actual
   PCMCIA card corresponds to one device instance, and is described
   by one dev_link_t structure (defined in ds.h).

   You may not want to use a linked list for this -- for example, the
   memory card driver uses an array of dev_link_t pointers, where minor
   device numbers are used to derive the corresponding array index.
*/

static dev_link_t *dev_list = NULL;
  

/*====================================================================*/


/*  specify our method entry points. struct defined in linux/fs.h */
struct file_operations das16s_fops = {
    NULL,           /* lseek */
    das16s_fop_read,
    NULL,           /* write */
    NULL,           /* readdir */
    NULL,           /* poll */
    das16s_fop_ioctl,
    NULL,           /* mmap */
    das16s_fop_open,
    NULL,           /* flush */
    das16s_fop_release,
    NULL,           /* fsync */
    NULL,           /* fasync */
    NULL,           /* check_media_change */
    NULL,           /* revalidate */    
    NULL            /* lock */    
};

static int das16s_major = 0; /* dynamic major device number by default */
static int testing=0;
static unsigned char in_fop_ioctl = FALSE;                          /* for testing only ?!? */
static unsigned char in_irq_handler=FALSE;                                           /* reentrance flag */
static unsigned long burst_in_process=FALSE;


/*******************************************/
static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}


/*======================================================================

    das16s_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.
    
======================================================================*/
/*******************************************/
static dev_link_t *das16s_attach(void)
{
    das16s_dev_t *local;
    dev_link_t *link;
    client_reg_t client_reg;
    int ret, i;
    
    DEBUG(3, "das16s_attach()\n");

    /* Allocate space for private device-specific data (and the link struct) */
    local = kmalloc(sizeof(das16s_dev_t), GFP_KERNEL);
    if (!local)
    {
        DEBUG(0, "bad kmalloc of local info struct!\n");        
        return NULL;
    }
    DEBUG(3,"das16s_attach():local addr = %p\n",local);    /* ?!? for testing*/
    memset(local, 0, sizeof(das16s_dev_t));
    link = &local->link;
    link->priv = local;
    DEBUG(3,"das16s_attach():link addr = %p\n",link);    /* ?!? for testing*/
    DEBUG(3,"das16s_attach():link->priv addr = %p\n",link->priv);    /* ?!? for testing*/
    
    /* Initialize the dev_link_t structure */
    link->release.function = &das16s_release;
    link->release.data = (u_long)link;

    /* Interrupt setup */
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    if (irq_list[0] == -1)
	link->irq.IRQInfo2 = irq_mask;
    else
	for (i = 0; i < 4; i++)
	    link->irq.IRQInfo2 |= 1 << irq_list[i];
    link->irq.Handler = NULL;
    
    /*
      General socket configuration defaults can go here.  In this
      client, we assume very little, and rely on the CIS for almost
      everything.  In most clients, many details (i.e., number, sizes,
      and attributes of IO windows) are fixed by the nature of the
      device, and can be hard-wired here.
    */
    link->conf.Attributes = 0;
    link->conf.Vcc = 50;
    link->conf.IntType = INT_MEMORY_AND_IO;

    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    DEBUG(3, "das16s_attach(): dev_list addr = %p\n",dev_list);    /* ?!? for testing*/
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =
	CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
	CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
	CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &das16s_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;

    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != CS_SUCCESS) {
	cs_error(link->handle, RegisterClient, ret);
	das16s_detach(link);
	return NULL;
    }

    DEBUG(3, "das16s_attach() done.\n");
    return link;
} /* das16s_attach */


/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/
/*******************************************/
static void das16s_detach(dev_link_t *link)
{
    dev_link_t **linkp;

    DEBUG(3, "das16s_detach(0x%p)\n", link);
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    /*
       If the device is currently configured and active, we won't
       actually delete it yet.  Instead, it is marked so that when
       the release() function is called, that will trigger a proper
       detach().
    */
    if (link->state & DEV_CONFIG) {
    	DEBUG(2,"das16s: detach postponed, '%s' "
	       "still locked\n", link->dev->dev_name);
	link->state |= DEV_STALE_LINK;
	return;
    }

    /* Break the link with Card Services */
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, and free it */
    *linkp = link->next;
    /* This points to the parent das16s_dev_t struct */
    kfree(link->priv);
    
} /* das16s_detach */


/*======================================================================

    das16s_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    device available to the system.
    
======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed

#define CFG_CHECK(fn, args...) \
if (CardServices(fn, args) != 0) goto next_entry
/*******************************************/
static void das16s_config(dev_link_t *link)
{
    client_handle_t handle = link->handle;
    das16s_dev_t *dev = link->priv;
    tuple_t tuple;
    cisparse_t parse;
    int last_fn, last_ret, ret, i/*, j*/;
    u_char buf[64];
    config_info_t conf;
    win_req_t req;
    memreq_t map;
    conf_reg_t cis_conf;
    /*int i;*/  /* ?!? for testing*/
    
    DEBUG(3, "das16s_config(0x%p)\n", link);

    /*
       This reads the card's CONFIG tuple to find its configuration
       registers.
    */
    tuple.DesiredTuple = CISTPL_CONFIG;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];
    
    /* Configure card */
    link->state |= DEV_CONFIG;

    /* Look up the current Vcc */
    CS_CHECK(GetConfigurationInfo, handle, &conf);
    link->conf.Vcc = conf.Vcc;

    /*
      In this loop, we scan the CIS for configuration table entries,
      each of which describes a valid card configuration, including
      voltage, IO window, memory window, and interrupt settings.

      We make no assumptions about the card to be configured: we use
      just the information available in the CIS.  In an ideal world,
      this would work for any PCMCIA card, but it requires a complete
      and accurate CIS.  In practice, a driver usually "knows" most of
      these things without consulting the CIS, and most client drivers
      will only use the CIS to fill in implementation-defined details.
    */
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    while (1) {
	cistpl_cftable_entry_t dflt = { 0 };
	cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
	CFG_CHECK(GetTupleData, handle, &tuple);
	CFG_CHECK(ParseTuple, handle, &tuple, &parse);

	if (cfg->flags & CISTPL_CFTABLE_DEFAULT) dflt = *cfg;
	if (cfg->index == 0) goto next_entry;
	link->conf.ConfigIndex = cfg->index;
	
	/* Does this card need audio output? */
	if (cfg->flags & CISTPL_CFTABLE_AUDIO) {
	    link->conf.Attributes |= CONF_ENABLE_SPKR;
	    link->conf.Status = CCSR_AUDIO_ENA;
	}
	
	/* Use power settings for Vcc and Vpp if present */
	/*  Note that the CIS values need to be rescaled */
	if (cfg->vcc.present & (1<<CISTPL_POWER_VNOM)) {
	    if (conf.Vcc != cfg->vcc.param[CISTPL_POWER_VNOM]/10000)
		goto next_entry;
	} else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM)) {
	    if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM]/10000)
		goto next_entry;
	}
	    
	if (cfg->vpp1.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vpp1 = link->conf.Vpp2 =
		cfg->vpp1.param[CISTPL_POWER_VNOM]/10000;
	else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
	    link->conf.Vpp1 = link->conf.Vpp2 =
		dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
	
	/* Do we need to allocate an interrupt? */
	if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1)
	    link->conf.Attributes |= CONF_ENABLE_IRQ;
	
	/* IO window settings */
	link->io.NumPorts1 = link->io.NumPorts2 = 0;
	if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0)) {
	    cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
	    link->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
	    if (!(io->flags & CISTPL_IO_8BIT))
            {
		link->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
                DEBUG(3, "das16s_config():io data path is 16 wide\n");
            }
	    if (!(io->flags & CISTPL_IO_16BIT))
            {
		link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
                DEBUG(3, "das16s_config():io data path is 8 wide\n");
            }
	    link->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
	    link->io.BasePort1 = io->win[0].base;
	    link->io.NumPorts1 = io->win[0].len;
	    if (io->nwin > 1) {
		link->io.Attributes2 = link->io.Attributes1;
		link->io.BasePort2 = io->win[1].base;
		link->io.NumPorts2 = io->win[1].len;
	    }
	}

	/* This reserves IO space but doesn't actually enable it */
	CFG_CHECK(RequestIO, link->handle, &link->io);

	/*
	  Now set up a common memory window, if needed.  There is room
	  in the dev_link_t structure for one memory window handle,
	  but if the base addresses need to be saved, or if multiple
	  windows are needed, the info should go in the private data
	  structure for this device.

	  Note that the memory window base is a physical address, and
	  needs to be mapped to virtual space with ioremap() before it
	  is used.
	*/
	if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0)) {
	    cistpl_mem_t *mem =
		(cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
	    req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
	    req.Attributes |= WIN_ENABLE;
	    req.Base = mem->win[0].host_addr;
	    req.Size = mem->win[0].len;
	    req.AccessSpeed = 0;
	    link->win = (window_handle_t)link->handle;
	    CFG_CHECK(RequestWindow, &link->win, &req);
	    map.Page = 0; map.CardOffset = mem->win[0].card_addr;
	    CFG_CHECK(MapMemPage, link->win, &map);
	}
	/* If we got this far, we're cool! */
	break;
	
    next_entry:
	CS_CHECK(GetNextTuple, handle, &tuple);
    }
    
    /*
       Allocate an interrupt line.  Note that this does not assign a
       handler to the interrupt, unless the 'Handler' member of the
       irq structure is initialized.
    */
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
	CS_CHECK(RequestIRQ, link->handle, &link->irq);
	
    /*
       This actually configures the PCMCIA socket -- setting up
       the I/O windows and the interrupt mapping, and putting the
       card and host interface into "Memory and IO" mode.
    */
    CS_CHECK(RequestConfiguration, link->handle, &link->conf);

    /*
      We can release the IO port allocations here, if some other
      driver for the card is going to loaded, and will expect the
      ports to be available.
    */
    if (free_ports) {
	if (link->io.BasePort1)
	    release_region(link->io.BasePort1, link->io.NumPorts1);
	if (link->io.BasePort2)
	    release_region(link->io.BasePort2, link->io.NumPorts2);
    }

    /*
      At this point, the dev_node_t structure(s) need to be
      initialized and arranged in a linked list at link->dev.
    */
    sprintf(dev->node.dev_name, "das16s");
    dev->node.major = dev->node.minor = 0;
    link->dev = &dev->node;

    /* Finally, report what we've done */
    printk(KERN_DEBUG "%s: index 0x%02x: Vcc %d.%d",
	   dev->node.dev_name, link->conf.ConfigIndex,
	   link->conf.Vcc/10, link->conf.Vcc%10);
    if (link->conf.Vpp1)
	printk(", Vpp %d.%d", link->conf.Vpp1/10, link->conf.Vpp1%10);
    if (link->conf.Attributes & CONF_ENABLE_IRQ)
	printk(", irq %d", link->irq.AssignedIRQ);
    if (link->io.NumPorts1)
	printk(", io 0x%04x-0x%04x", link->io.BasePort1,
	       link->io.BasePort1+link->io.NumPorts1-1);
    if (link->io.NumPorts2)
	printk(" & 0x%04x-0x%04x", link->io.BasePort2,
	       link->io.BasePort2+link->io.NumPorts2-1);
    if (link->win)
	printk(", mem 0x%06lx-0x%06lx", req.Base,
	       req.Base+req.Size-1);
    printk("\n");

    cis_conf.Action = CS_READ;                                              /* read these regs just for fun -SR  ?!? */
    cis_conf.Offset = CISREG_CCSR;
    if (CardServices(AccessConfigurationRegister, link->handle, &cis_conf) != 0)
        DEBUG(1, "Error return from Cardservices\n");
    DEBUG(3,"das16s: CCSR is %u, ",cis_conf.Value);
    cis_conf.Action = CS_READ;
    cis_conf.Offset = CISREG_COR;
    if (CardServices(AccessConfigurationRegister, link->handle, &cis_conf) != 0)
        DEBUG(1, "Error return from Cardservices\n");
    DEBUG(3,"COR is %u\n",cis_conf.Value);

    link->state &= ~DEV_CONFIG_PENDING;

    dev->base_address = link->io.BasePort1;                             /* capture io address of card*/
    dev->irq = link->irq.AssignedIRQ;
    dev->adc_rate_divisor = 10;
    dev->aux_adc_rate_divisor = 10;
    dev->channel_scan = MAX_CHANNELS;
    dev->volt_range = BI_TEN_VOLT << 8;                                     /* safest default */
    dev->start = FALSE;
    dev->num_mux_banks = 1;
    dev->burst_acq = FALSE;
    dev->num_dig_in_banks =1;
    dev->dio_output_val = (u16)init_dio /*<< 8*/;                                     /* capture module's initial dio value */
/*    for (i=0; i < MAX_BANKS; i++)
        dev->old_digital_input_word[i] = 85;*/                                       /* different from output val */
    dev->digital_input_word = 0;
    dev->multiplexing = FALSE;
    dev->send_dig_ins = TRUE;
    dev->chan_cnt = 0;                                                  /* havent gotten any samples yet */
    dev->mux = 0;                                                           /* havent gotten any samples yet */
    /*dev->sample_bank[HDR_MAGIC_OFFSET] = SAMPLE_HEADER_MAGIC_1;*/             /* identifies beginning of result packets in a fifo */
    /*dev->sample_bank[HDR_MAGIC_OFFSET+1] = SAMPLE_HEADER_MAGIC_2;*/             /* identifies beginning of result packets in a fifo */
    dev->error_correcting = FALSE;
    dev->dio_lower_dir = DIO_OUT << 6;                                      /* both nibbles do output at first */
    dev->dio_upper_dir = DIO_OUT << 7;
    /*dev->ignore_next_bank = FALSE;*/                                            /* first bank of samples is valid */
    dev->scan_begin = SCAN_BEGIN_0;
    dev->scan_end = SCAN_END_15;
    dev->result_table_head = 0;                                             /* init to begin of results table */
    dev->result_table_tail = 0;                                             /* init to begin of results table */
    dev->result_pack_avail = 0;                                             /* no packets available in table at start */
    for (i=0; i < MAX_SAMPLE_PACKETS+1; i++)                                /* fill result table magic fields */
    {
        /*for (j=ADC_DATA_OFFSET; j < ADC_DATA_QTY+ADC_DATA_OFFSET; j++)  
            dev->result_table[i][j] = i;*/  /* for testing only ?!? */
        dev->result_table[i][0] = SAMPLE_HEADER_MAGIC_1;
        dev->result_table[i][1] = SAMPLE_HEADER_MAGIC_2;
    }
    /*dev->result_pack_avail = MAX_SAMPLE_PACKETS;*/    /* just for testing !!! */

    ret = rtl_request_irq(dev->irq, das16s_irq_handler);           /* install our handler with RTLinux executive */
    if (ret)                                                            /* nonzero return means problems */
    {
            printk(KERN_WARNING "das16s: ERROR installing RT interrupt handler: !\n");
            if (ret == -EBUSY)
                printk("das16s: RT-Interrupt is already in use.");
            else if (ret == -EINVAL)
                printk("das16s: Invalid interrupt number, or null handler.");
    }
    else
    {
        rtl_hard_enable_irq(dev->irq);                                  /* enable this level at mask register */
        DEBUG(3, "das16s_config(): requested and hard enabled irq %d w/ handler %p\n",dev->irq,das16s_irq_handler);
    }
    /*DEBUG(1,"config(): dev addr = %p\n",dev); */   /* ?!? for testing */
    /*DEBUG(1,"config(): card base addr = 0x%x\n",dev->base_address); */   /* ?!? for testing*/
    if (testing)
        DEBUG(1,"das16s: Driver is in testing mode !\n");
    DEBUG(3, "das16s_config() done.\n");

    return;

cs_failed:
    cs_error(link->handle, last_fn, last_ret);
    das16s_release((u_long)link);

} /* das16s_config */


/*======================================================================

    After a card is removed, das16s_release() will unregister the
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/
/*******************************************/
static void das16s_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;

    DEBUG(0, "das16s_release(0x%p)\n", link);

    /*
       If the device is currently in use, we won't release until it
       is actually closed, because until then, we can't be sure that
       no one will try to access the device or its data structures.
    */
    if (link->open) {
	DEBUG(1, "das16s: release postponed, '%s' still open\n",
	      link->dev->dev_name);
	link->state |= DEV_STALE_CONFIG;
	return;
    }

    /* FIXME: any point writing to the card here ?!? */
    outw(0,link->io.BasePort1+DAS_STATUS_REG);     /* stop interrupts by specifying no Int source and resetting int enable bit */

    /* Unlink the device chain */
    link->dev = NULL;

    /*
      In a normal driver, additional code may be needed to release
      other kernel data structures associated with this device. 
    */

    
    rtl_free_irq(link->irq.AssignedIRQ);                                   /* unregister our handler */
    
    /* Don't bother checking to see if these succeed or not */
    if (link->win)
	CardServices(ReleaseWindow, link->win);
    CardServices(ReleaseConfiguration, link->handle);
    if (link->io.NumPorts1)
	CardServices(ReleaseIO, link->handle, &link->io);
    if (link->irq.AssignedIRQ)
	CardServices(ReleaseIRQ, link->handle, &link->irq);
    link->state &= ~DEV_CONFIG;
    
    if (link->state & DEV_STALE_LINK)
	das16s_detach(link);
    
} /* das16s_release */


/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.

    When a CARD_REMOVAL event is received, we immediately set a
    private flag to block future accesses to this device.  All the
    functions that actually access the device should check this flag
    to make sure the card is still present.
    
======================================================================*/
/*******************************************/
static int das16s_event(event_t event, int priority,
		       event_callback_args_t *args)
{
    dev_link_t *link = args->client_data;
    das16s_dev_t *dev = link->priv;
    
    DEBUG(3, "das16s_event(0x%06x)\n", event);
    
    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
        DEBUG(3, "das16s_event REMOVAL\n");
	link->state &= ~DEV_PRESENT;
	if (link->state & DEV_CONFIG) {
	    ((das16s_dev_t *)link->priv)->stop = 1;
	    link->release.expires = jiffies + HZ/20;
	    add_timer(&link->release);
	}
        dev->start = FALSE;                                             /* mark device as stopped */
	break;
    case CS_EVENT_CARD_INSERTION:
        DEBUG(3, "das16s_event INSERTION\n");
	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	dev->bus = args->bus;
	das16s_config(link);
	break;
    case CS_EVENT_PM_SUSPEND:
        DEBUG(3, "das16s_event SUSPEND\n");
	link->state |= DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
        DEBUG(3, "das16s_event RESET_PHYS\n");
	/* Mark the device as stopped, to block IO until later */
	dev->stop = 1;
	if (link->state & DEV_CONFIG)
	    CardServices(ReleaseConfiguration, link->handle);
        dev->start = FALSE;                                       /* mark device as stopped */
	break;
    case CS_EVENT_PM_RESUME:
        DEBUG(3, "das16s_event PM_RESUME\n");
	link->state &= ~DEV_SUSPEND;
	/* Fall through... */
    case CS_EVENT_CARD_RESET:
        DEBUG(3, "das16s_event RESET\n");
	if (link->state & DEV_CONFIG)
        {
	    CardServices(RequestConfiguration, link->handle, &link->conf);
            das16s_reset();
        }
        dev->start = TRUE;                                       /* mark device as started */
	break;
    }
    return 0;
} /* das16s_event */


/* this is effectively init_module() */
/*====================================================================*/
/*******************************************/
static int __init init_das16s(void)
{
    servinfo_t serv;
    int ret;

    DEBUG(0, "%s\n", version);
    DEBUG(3, "das16s: init_das16s()\n");
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) {
	printk(KERN_NOTICE "das16s: Card Services release "
	       "does not match!\n");
	return -1;
    }
    register_pccard_driver(&dev_info, &das16s_attach, &das16s_detach);
    /*printk("das16s: finished registering\n");*/

#if 0
    ret = rtf_create(RESULT_FIFO_ID, RESULT_FIFO_LEN);                      /* create (outgoing) ADC results fifo */
    if (ret != 0)                                                           /* problem creating fifo */
    {
        printk("Error creating rtfifo %d\n",RESULT_FIFO_ID);
        if (ret == -ENODEV)
            printk("fifo number is too great.\n");
        else if (ret == -EBUSY)
            printk("fifo already in use.\n");
        else if (ret == -ENOMEM)
            printk("memory could not be allocated for fifo.\n");
        else
            printk("return value: %d\n",ret);
    }

    ret = rtf_create(CTRL_FIFO_ID, CTRL_FIFO_LEN);                          /* create (incoming) control fifo */
    if (ret != 0)                                                           /* problem creating fifo */
    {
        printk("Error creating rtfifo %d\n",CTRL_FIFO_ID);
        if (ret == -ENODEV)
            printk("fifo number is too great.\n");
        else if (ret == -EBUSY)
            printk("fifo already in use.\n");
        else if (ret == -ENOMEM)
            printk("memory could not be allocated for fifo.\n");
        else
            printk("return value: %d\n",ret);
    }
    ret = rtf_create_handler(CTRL_FIFO_ID, das16s_ctl_fifo_handler);                        /* register our handler - invoked when fifo is written */
    if (ret)
        printk("Error creating handler for rtfifo %d: %d",CTRL_FIFO_ID,ret);

    rtf_create(DIG_RESULT_FIFO_ID, DIG_RESULT_FIFO_LEN);                        /* create (outgoing) digital input result fifo */
    if (ret)
        printk("Error creating rtfifo %d: %d",DIG_RESULT_FIFO_ID,ret);

    printk(KERN_INFO "Result rt-fifo: %d, config rt-fifo: %d, dig-input result rt-fifo: %d\n",
            RESULT_FIFO_ID, CTRL_FIFO_ID,DIG_RESULT_FIFO_ID);
#endif

    /* Register device major number, accepting a dynamic number. */
    /* read /proc/devices after to find what major was assigned */
    ret = register_chrdev(das16s_major, dev_info, &das16s_fops);
    if (ret < 0) {
        printk(KERN_WARNING "das16s: can't get major %d\n",das16s_major);
        /*return ret;?!?*/
    }
    else {
        das16s_major = ret;
        printk(KERN_INFO "das16s: device major number is %d\n",das16s_major);
    }

    return 0;
}


/*******************************************/
static void __exit exit_das16s(void)
{
    DEBUG(3, "das16s: exit_das16s()\n");

    unregister_pccard_driver(&dev_info);
    while (dev_list != NULL) {
	if (dev_list->state & DEV_CONFIG)
	    das16s_release((u_long)dev_list);
	das16s_detach(dev_list);
    }

    unregister_chrdev(das16s_major,dev_info);
}

module_init(init_das16s);                                              /* register our function to be called for module initialization */
module_exit(exit_das16s);                                                    /* register our function to be called for module shutdown */


/*******************************************/
int das16s_reset(void)
{
    das16s_dev_t *dev;                                                        /* to access das16s struct */
    ioaddr_t base_address;                                                    /* for base address of das16s */

    DEBUG(0, "das16s_reset()\n");

    if (dev_list == NULL)                                                     /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "das16s: modlule has not been initialized properly!");
        return -1;
    }
    else
    {
        dev = (das16s_dev_t *) dev_list->priv;                                /* get addr of our private struct */
        base_address = dev->base_address;                                     /* get io addr of card */
    }

#if 0
    /* program all 3 timers just for a functional check - works OK */
    /* inital card config: read the OR'd constants */
    outw(GAIN_BI_10V|DIO_UPR_OUTPUT|DIO_LWR_OUTPUT|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,base_address+DAS_CTRL_REG);

    /* configure ADC timer 0; load LSByte then MSB; square wave generator; binary counts */
    outw(COUNTER_0|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN/*0x3E*/,base_address+DAS_CTR_CTL_REG);
    /* external ctr out will be greatest possible period = .65535sec */
    outw(0xFFFF & 0xff,base_address+DAS_CTR0_DATA_REG);          /* write LSB for counter */
    outw(0xFFFF >> 8,base_address+DAS_CTR0_DATA_REG);            /* write MSB for counter */
    WAIT_A_WHILE(1);                                             /* wait for regs to settle here*/

    /* configure ADC timer 1; load LSByte then MSB; square wave generator; binary counts */
    outw(COUNTER_1|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN/*0x7E*/,base_address+DAS_CTR_CTL_REG);
    outw(10 & 0xff,base_address+DAS_CTR1_DATA_REG);          /* write LSB for counter */
    outw(0 >> 8,base_address+DAS_CTR1_DATA_REG);            /* write MSB for counter */
    WAIT_A_WHILE(1);                                                     /* wait for regs to settle here*/

    /* configure ADC timer 2; load LSByte then MSB; square wave generator; binary counts */
    outw(COUNTER_2|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN/*0xBE*/,base_address+DAS_CTR_CTL_REG);
    outw(100 & 0xff,base_address+DAS_CTR2_DATA_REG);          /* write LSB for counter */
    outw(0 >> 8,base_address+DAS_CTR2_DATA_REG);            /* write MSB for counter */
    WAIT_A_WHILE(1);                                                     /* wait for regs to settle here*/
#endif

#if 0   
    /* test of external irqs - works OK */
    /* stop ints by specifying no int source, soft trig, and reset int enable bit */
    outw(0,dev->base_address+DAS_STATUS_REG); 

    /* inital card config: read the OR'd constants */
    outw(GAIN_BI_10V|DIO_LWR_OUTPUT|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,dev->base_address+DAS_CTRL_REG);

    /* enable ints */
    outw(ENAB_INT|INT_SRC_EXTERN,dev->base_address+DAS_STATUS_REG);

    outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);
    outw(ENAB_INT|INT_SRC_EXTERN,dev->base_address+DAS_STATUS_REG);
#endif

#if 0
    /* test of irq generation by ctr2 - works OK */
    /* stop ints by specifying no int source, soft trig, and reset int enable bit */
    outw(0,dev->base_address+DAS_STATUS_REG);

    /* inital card config: read the OR'd constants */
    outw(GAIN_BI_10V|DIO_LWR_OUTPUT|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,dev->base_address+DAS_CTRL_REG);

    /* enable ints */
    outw(ENAB_INT|INT_SRC_CTR_2,dev->base_address+DAS_STATUS_REG);

    outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);
    outw(ENAB_INT|INT_SRC_CTR_2,dev->base_address+DAS_STATUS_REG);

    /* write modes for counters 2 and 1 */
    outw(COUNTER_2|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN,dev->base_address+DAS_CTR_CTL_REG);
    WAIT_A_WHILE(1);    
    outw(COUNTER_1|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN,dev->base_address+DAS_CTR_CTL_REG);
    WAIT_A_WHILE(1);

    /* write divisor for counters 2 and 1 */
    outw(65535/*adc_divisor*/ & 0xff,dev->base_address+DAS_CTR2_DATA_REG);          /* write LSB for counter */
    WAIT_A_WHILE(1);                                     /*?!? wait for regs to settle here*/
    outw(65535/*adc_divisor*/ >> 8,dev->base_address+DAS_CTR2_DATA_REG);            /* write MSB for counter */
    WAIT_A_WHILE(1); /*?!? wait for regs to settle here*/
    outw(100 & 0xff,dev->base_address+DAS_CTR1_DATA_REG);                                 /* write LSB */
    WAIT_A_WHILE(1);                                     /*?!? wait for regs to settle here*/
    outw(0 >> 8,dev->base_address+DAS_CTR1_DATA_REG);                                    /* write MSB */
    WAIT_A_WHILE(1);
#endif

    DEBUG(1,"das16s_reset(): done\n");

    return 0;
} /* das16s_reset */


/*******************************************/
int das16s_fop_open(struct inode *inode_p, struct file *file_p)
{
    DEBUG(3,"das16s: fop_open()\n");
    /*in fs.h: MODE_READ 1, FMODE_WRITE 2*/
    /*printk(KERN_INFO "fop_open(): f_mode is 0x%x\n",file_p->f_mode);*/
    MOD_INC_USE_COUNT;

    return 0;
}


/*******************************************/
ssize_t das16s_fop_read(struct file *filp, char *buf, size_t count, loff_t *offset_p)
{
    struct inode* inode = filp->f_dentry->d_inode;
    /*unsigned int minor = MINOR(inode->i_rdev);*/
    char data[13]={"Hello Folks!"};
    char *data_p;
    int packets_read=0, packet_count=0;
    das16s_dev_t *dev;                                                        /* to access das16s struct */
    unsigned int interrupt_state=0;

    if (dev_list == NULL) {                                                   /* if there is a list of instances of the card */
        printk(KERN_WARNING "das16s_fop_read(): module has not been initialized properly!");
        return ERROR;
    }
    else
        dev = (das16s_dev_t *) dev_list->priv;

    DEBUG(3, "das16s: fop_read(): count is %d, buf addr is %p\n", count, buf);

    if (filp->f_flags & O_NONBLOCK)                /* if reader is nonblocking */
    {
        if (dev->result_pack_avail == 0)   /* no data avail in buffer */
        {
            DEBUG(3,"no results, returning 0\n");
            return 0;   /* ?!? -E */
        }
    }
    else                                          /* reader will block */
    {
    #if 0
        /* FIXME: implement ?!? */
		while (RTF_EMPTY(minor)  )  /* loop until data */
                {
			if (signal_pending(current))
				return -ERESTARTSYS;
			fifo_setup_sleep(minor);
			interruptible_sleep_on_timeout(&RTF_WAIT(minor),
					RTL_FIFO_TIMEOUT);
		}
    #endif
    }
                    
    data_p = data;
    /*printk("our data=%s\n",data_p);*/

    packet_count = count/RESULT_PACK_SIZE_B;                                      /* reduce to count of packets to write */
    /*printk("packet count: %d\n",packet_count);*/
/*    if (packet_count > MAX_SAMPLE_PACKETS)
        packet_count = MAX_SAMPLE_PACKETS; */
    if (packet_count > dev->result_pack_avail)                         /* user may only have as many result packets as card fifo holds */
        packet_count = dev->result_pack_avail;
    DEBUG(3,"trimmed packet count: %d\n",packet_count);

    while (packet_count > 0)
    {
        copy_to_user(buf, GET_ADC_RESULT_PACKET(dev), RESULT_PACK_SIZE_B);  /* actually write data to user */
        rtl_critical(interrupt_state);                                      /* mask irqs so we don't get preempted ? */
        CLEAR_ADC_RESULT(dev);                                              /* decrement table qty by one */
        rtl_end_critical(interrupt_state);            
        buf += RESULT_PACK_SIZE_B;                                          /* movbe forward in user buffer */
        packet_count -= 1;
	packets_read += 1;                                  /* increment count of how many packets user has read */
    }

    if (packets_read) {                             /* if we wrote any */
        inode->i_atime = CURRENT_TIME;
/*		if ((result = (*RTF_HANDLER(minor)) (minor)) < 0) {
			return result;
		}*/
            /*printk("fop_read(): returning %d\n",read);*/
	return (packets_read*RESULT_PACK_SIZE_B);   /* qty of bytes written */
    }

	return 0;
}   /* das16s_fop_read */


/* FIXME: phase out arg and store global config on ioctl */
/* note that magic numbers have not been assigned to cmd, so errant ioctls could reconfigure this card */
/*******************************************/
int das16s_fop_ioctl(struct inode *inode_p, struct file *file_p, unsigned int cmd, unsigned long arg)
{
    int ret;
    ctl_msg_struct_t *ctl_p;                                                /* this struct sent to us from user, holds configuration */
    das16s_dev_t *dev;                                                        /* to access das16s struct */
    /*unsigned short bitmask;*/
    static int last_command=-1;

    if (testing)
    {
        in_fop_ioctl = TRUE;                                                /* setglobal flag to test for errant calls during irq */
    }

    DEBUG(3, "das16s: fop_ioctl(): cmd is %d, arg is %p\n",cmd,(void *)arg);    

    if (dev_list == NULL)                                                     /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "das16s_fop_ioctl(): module has not been initialized properly!");
        return ERROR;
    }
    else
        dev = (das16s_dev_t *) dev_list->priv;

    if (testing)
        if (in_irq_handler)
            printk("das16s Error: in fop_ioctl() during irq handler!\n");       /* should never happen */

    /* see if pointer is valid. defined in uaccess.h */
    ret = verify_area(VERIFY_READ,(ctl_msg_struct_t *)arg,sizeof(ctl_msg_struct_t));
    if (ret != 0)                                                           /* problem */
    {
        printk("das16s_fop_ioctl(): Couldn't read user struct!\n");
        return ret;
    }
    ctl_p = (ctl_msg_struct_t *)arg;

    switch (ctl_p->conversion_mode)
    {
    case CONTIN_ACQUISITION:
        dev->burst_acq = FALSE;                                   /* allow acquisition to be continuous */
        DEBUG(3, "set contin acq\n");
        break;
    case BURST_ACQUISITION:
        dev->burst_acq = TRUE;                                    /* do a burst of all mux banks and disable counters */
        DEBUG(3, "set burst acq\n");
        break;
    }

    switch (cmd)
    {
    case CONFIGURE_ACQUISITION:                              /* configure the accquisition mode */
        if ((ctl_p->aux_adc_rate_divisor > 0) && (ctl_p->adc_rate_divisor > 0))
            if ((INTERNAL_CLOCK_HZ/ctl_p->aux_adc_rate_divisor)/ctl_p->adc_rate_divisor <= MAX_SAMPLE_RATE_HZ)
            {
                dev->adc_rate_divisor = ctl_p->adc_rate_divisor;            /* set counter divisor, start acq uses this */
                dev->aux_adc_rate_divisor = ctl_p->aux_adc_rate_divisor;            /* set counter divisor, start acq uses this */
                DEBUG(3, "rate div: %u, aux rate div: %u\n",dev->adc_rate_divisor,dev->aux_adc_rate_divisor);
            }
        if ((ctl_p->num_mux_banks <= MAX_BANKS) && (ctl_p->num_mux_banks > 0))
        {
            dev->num_mux_banks = ctl_p->num_mux_banks;               /* set number of banks for the card to multiplex */
            DEBUG(3, "num mux banks: %d\n",dev->num_mux_banks);
        }
        else
            dev->num_mux_banks = 1;                                /* default is one bank */
        if ((ctl_p->num_chan > 0) && (ctl_p->num_chan <= MAX_CHANNELS))   /* can only sample 1-16 channels on this card */
        {
            dev->channel_scan = ctl_p->num_chan;            /* set number of channels for the card to scan (1-16) */
            dev->chan_cnt = 0;                                                /* re-init num samples gotten */
            DEBUG(3, "num chan: %d\n",dev->channel_scan);
        }
        if ((ctl_p->volt_range > -1) && (ctl_p->volt_range < 9))      /* any single ended or bopolar 10 volt, 5 volt. 2.5 volt. */
        {
            dev->volt_range = (ctl_p->volt_range << 8);            /* clear the two lsb's of input mode, "0" is 10V range */
            DEBUG(3, "volt range code : %d\n",dev->volt_range);
        }
        if (ctl_p->error_correcting > -1)                         /* if user wants to use calibration constants on results */
        {
            dev->error_correcting = ctl_p->error_correcting;
            DEBUG(3,"error correction: %d\n",dev->error_correcting);
        }
        if ((ctl_p->num_dig_in_banks > 0) && (ctl_p->num_dig_in_banks < 5))
        {
            dev->num_dig_in_banks = ctl_p->num_dig_in_banks;
            DEBUG(3, "got num dig in banks: %d\n",ctl_p->num_dig_in_banks);
        }
        else
            dev->num_dig_in_banks = 0;
        dev->send_dig_ins = TRUE;  /*?!? necessary on this card ?*/           /* this change forces a new digital sample */
        if ((dev->num_mux_banks > 1) || (dev->num_dig_in_banks > 1))          /* if any multi-bank requests came in */
        {
            dev->multiplexing = TRUE;
            dev->mux = 0;                                                   /* re-init */
            DEBUG(3, "das16s: multiplexing.\n");
        }
        else
            dev->multiplexing = FALSE;
        if (dev->multiplexing)                                          /* if we must multiplex... */
        {
            dev->dio_upper_dir = DIO_OUT << 7;                          /* upper nibble is dig out */
            if (dev->num_dig_in_banks == 0)                             /* if no dig ins wanted ... */
                dev->dio_lower_dir = DIO_OUT << 6;                      /* lower nibble is also dig out */
            else                                                        /* if any dig ins wanted, lower nibble is dig in */
                dev->dio_lower_dir = DIO_IN << 6;
        }
        else                                                            /* not multiplexing */
        {
            if (dev->num_dig_in_banks == 0)                             /* if we dont want any dig ins */
            {
                dev->dio_upper_dir = DIO_OUT << 7;                      /* both nibbles are dig out */
                dev->dio_lower_dir = DIO_OUT << 6;
            }
            else                                                        /* we want one bank of dig in, and 0 or 1 bank of adc */
            {
                dev->dio_lower_dir = DIO_IN << 6;                       /* lower nibble digital in */
                dev->dio_upper_dir = DIO_OUT << 7;                      /* this one arbitrary choice is hopefully useful to people */
            }
        }
        das16s_program_adc();                                                /* now program the adc scan and counters */

        break;
    case START_ACQUISITION:                                                                 /* start accquisition */
    if (dev->burst_acq)                                                 /* if in burst mode */
    {
        if (burst_in_process)                                             /* new job called before previous job finished */
        {
            DEBUG(0, "das16s_fop_ioctl(): New burst request came before previous burst finished ! Ignoring it...\n");
            return ERROR;
        }
        else
            burst_in_process=TRUE;                                                /* global flag we are handling current request */
    }

        /*printk ("START: mux:%d chan_cnt:%d head:%d tail:%d \n",dev->mux,dev->chan_cnt,dev->result_table_head,dev->result_table_tail);*/
        /* 1st try: works OK */
        DAS_CLEAR_FIFO(dev->base_address); /* new w/ 7th */                  /* clear the fifo by reading samples out of it */
        /* must configure channel scan when starting new bank (& write dig out value) */
        outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_SCAN_REG); 
        /*DAS_CLEAR_FIFO(dev->base_address);*/ /* new w/ 8th */                  /* clear the fifo by reading samples out of it */
        /*WAIT_A_WHILE(1);*/     /* new w/ 8th */
        das16s_program_timers(ctl_p->adc_rate_divisor,ctl_p->aux_adc_rate_divisor, DAS_RESTART); /* now restart conversion pacer */

        if (dev->burst_acq)  /* new 9th */                    /* re-enable irqs, specify int source... */
            outw(ENAB_INT|INT_SRC_END_SCAN|TRIG_SRC_CTR2,dev->base_address+DAS_STATUS_REG); /* enable ints */

#if 0   /* 2nd try: still has glitches */
        das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_RESTART);
        outw(0,dev->base_address+DAS_CONV_REG);    /* write start conversion reg. Clears FIFO overrun & puts a conversion in FIFO?!? */
        /* wait for 'fifo not empty' flag to set (from first conversion ?!?) */
        while (!(inw(dev->base_address+DAS_CTRL_REG) & FIFO_NOT_EMPTY_MASK))
            ;
        /* configure channel scan & write dig out value. mux==0 */
        outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_SCAN_REG); 
        WAIT_A_WHILE(1);                          /* ?!? waste some CPU cycles waiting for regs to settle here ?!? */
        DAS_CLEAR_FIFO(dev->base_address);                                        /* clear the fifo by reading samples out of it */
        bitmask = inw(dev->base_address+DAS_STATUS_REG);    /* ?!? this helps */  /* rewrite status reg */
        outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);   /* clear int by rewriting ctrl reg */
        outw(bitmask,dev->base_address+DAS_STATUS_REG);
#endif
#if 0   /* 3rd try: still has glitches */
    /* stop ints by specifying no int source, soft trig, and reset int enable bit */
    outw(0,dev->base_address+DAS_STATUS_REG);

    /* stop pacer counters (2 and 1) by writing mode ?!? */
    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_STOP);

    /* inital card config: read the OR'd constants */
    outw(dev->volt_range|dev->dio_lower_dir|dev->dio_upper_dir|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,dev->base_address+DAS_CTRL_REG);

    if (dev->channel_scan > 1)                                              /* user wants a scan of several channels */
        outw(ENAB_INT|INT_SRC_END_SCAN|TRIG_SRC_CTR2,dev->base_address+DAS_STATUS_REG); /* enable ints */
    else
        ; /* FIXME: interrupt on FFNE  ?!? */

    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_START);

    outw(0,dev->base_address+DAS_CONV_REG);          /* write start conversion reg. Clears FIFO overrun & puts a conversion in FIFO?!? */

    /* wait for 'fifo not empty' flag to set (from first conversion ?!?) */
    while (!(inw(dev->base_address+DAS_CTRL_REG) & FIFO_NOT_EMPTY_MASK))
        ;

    /* configure channel scan & write dig out value. mux==0 */
    outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_SCAN_REG); 
    /*WAIT_A_WHILE(dev->base_address,1000); */                     /* waste 1/2 million CPU cycles waiting for regs to settle here ?!? */
    WAIT_A_WHILE(1);                          /* ?!? waste some CPU cycles waiting for regs to settle here ?!? */

    DAS_CLEAR_FIFO(dev->base_address);                                        /* clear the fifo by reading samples out of it */

    bitmask = inw(dev->base_address+DAS_STATUS_REG);    /* ?!? this helps */  /* rewrite status reg */
    outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);   /* clear int by rewriting ctrl reg */
    outw(bitmask,dev->base_address+DAS_STATUS_REG);
#endif
#if 0   /* 4th try: still glitches */
        DAS_CLEAR_FIFO(dev->base_address);                                    /* clear the fifo by reading samples out of it */
        outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_SCAN_REG);
        WAIT_A_WHILE(1);                          /* ?!? waste some CPU cycles waiting for regs to settle here ?!? */
        das16s_program_timers(ctl_p->adc_rate_divisor,ctl_p->aux_adc_rate_divisor, DAS_RESTART); /* now restart conversion pacer */
#endif
#if 0   /* 5th try: still glitches */
        outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_SCAN_REG);
        /*WAIT_A_WHILE(dev->base_address,5);*/                          /* waste some CPU cycles waiting for regs to settle here ?!? */
        DAS_CLEAR_FIFO(dev->base_address);                                        /* clear the fifo by reading samples out of it */
        das16s_program_timers(ctl_p->adc_rate_divisor,ctl_p->aux_adc_rate_divisor, DAS_START); /* now restart conversion pacer */
#endif
#if 0   /* 6th workaround: glitches (rarely - once in 5 hours)  */
        if (dev->burst_acq)                      /* re-enable irqs, specify int source... */
            outw(ENAB_INT|INT_SRC_END_SCAN|TRIG_SRC_CTR2,dev->base_address+DAS_STATUS_REG); /* enable ints */
        else
            ;
        dev->dio_output_val &= (~0x03 << 12/*8*/);/* wrong,already shifted to upper byte?*//* clear 2 LSBs of MSByte, value ORed w/ mux*/
        dev->dio_output_val |= (dev->mux << 12/*8*/);                 /* combine dig out word w/ mux val */
        outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_SCAN_REG);
        das16s_program_timers(ctl_p->adc_rate_divisor,ctl_p->aux_adc_rate_divisor, DAS_START); /* now restart conversion pacer */
#endif
        break;
    case STOP_ACQUISITION:                                                                 /* stop accquisition */
        das16s_program_timers(ctl_p->adc_rate_divisor,ctl_p->aux_adc_rate_divisor, DAS_STOP);
        break;
    case DIGITAL_OUTPUT:                                                                 /* digital output */
        dev->dio_output_val = ctl_p->dio_output_val;  /* store the 8-bit value (still in LSByte) */
        DEBUG(3,"orig dig out val: %d\n",dev->dio_output_val);
        if (dev->multiplexing)
        {
            dev->dio_output_val <<= 2;                         /* shift our value left two to make room for mux bits */
            dev->dio_output_val &= ~0x03;                       /* clear 2 LSBs, value will be ORed w/ mux */
            dev->dio_output_val |= dev->mux;                 /* combine dig out word w/ mux val */
            DEBUG(3,"ORed out val w/ mux val: %d\n",dev->mux);
        }
        if (dev->dio_lower_dir == (DIO_IN << 6))                    /* doing dig IN on LSByte, so do dig OUT on MSByte */
        {
                dev->dio_output_val <<= 12;                         /* output goes to upper two of MSBs of word */
                DEBUG(3,"shifted dio by 12\n");
        }
        else                                                        /* whole byte is digital out */
            dev->dio_output_val <<= 8;                              /* shift to MSByte of word */
        DEBUG(3,"final dig out val: %d\n",dev->dio_output_val);
        outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_DIO_REG); /* write digout val, config channel scan*/
        break;        
    case 4: /* not used anymore*/                                       /* frequency adjustment */
        /* FIXME: move to config ?!? */
        das16s_program_timers(ctl_p->adc_rate_divisor,ctl_p->aux_adc_rate_divisor, DAS_START);
        break;
    case QUERY_ACQUISITION:  /* case 6 */
        DEBUG(2,"das16s: -query-\n");
        DEBUG(2,"mux=%d, chan_cnt=%d, dio_output_val=%d, last_command=%d, testing=%d\n",
            dev->mux,dev->chan_cnt,dev->dio_output_val,last_command,testing);
        break;    
    default:
        DEBUG(1, "das16s_fop_ioctl: bad command (%u)\n",cmd);
        return -EINVAL;
        break;
    }
    last_command = cmd;

    if (testing)
        in_fop_ioctl = FALSE;   /* for testing only ?!? */

    return 0;
}   /* das16s_fop_ioctl */


/*******************************************/
int das16s_fop_release(struct inode *inode_p, struct file *file_p)
{
    DEBUG(3,"das16s: fop_release()\n");
    MOD_DEC_USE_COUNT;
    return 0;
}


/* start, stop, restart adc pacer timer/counters. */
/* note that 'start' stores adc divisors, 'restart' uses stored values */
/*******************************************/
void das16s_program_timers(unsigned int adc_divisor,unsigned int aux_divisor, bool mode)
{
    das16s_dev_t *dev;                                                        /* to access das16s struct */
    static unsigned old_adc_divisor=10, old_aux_divisor=10;

    DEBUG(3,"das16s_program_timers() mode=%d\n",mode);

    if (dev_list == NULL)                                                     /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "das16s_program_timers(): module has not been initialized properly!");
        return;
    }
    else
        dev = (das16s_dev_t *) dev_list->priv;                                /* get addr of our private struct */

    DEBUG(3, "adc_div is %u, aux is %u, mode is %d\n",adc_divisor,aux_divisor,mode);

    /*  do always: write both (2 and 1) counter modes, this stops pacer clocks  */
    outw(COUNTER_2|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN,dev->base_address+DAS_CTR_CTL_REG);
    WAIT_A_WHILE(1);
    outw(COUNTER_1|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN,dev->base_address+DAS_CTR_CTL_REG);
    WAIT_A_WHILE(1);

    if (mode == DAS_START)
    {
        /* write divisor for counters 2 and 1 */
        outw(adc_divisor & 0xff,dev->base_address+DAS_CTR2_DATA_REG);          /* write LSB for counter */
        WAIT_A_WHILE(1);                                                     /* wait for regs to settle here*/
        outw(adc_divisor >> 8,dev->base_address+DAS_CTR2_DATA_REG);            /* write MSB for counter */
        WAIT_A_WHILE(1);                                                     /* wait for regs to settle here*/
        outw(aux_divisor & 0xff,dev->base_address+DAS_CTR1_DATA_REG);                                 /* write LSB */
        WAIT_A_WHILE(1);                                     /*?!? wait for regs to settle here*/
        outw(aux_divisor >> 8,dev->base_address+DAS_CTR1_DATA_REG);                                    /* write MSB */
        WAIT_A_WHILE(1);                                                     /* wait for regs to settle here */

        old_adc_divisor = adc_divisor;                                      /* save these values for restarts */
        old_aux_divisor = aux_divisor;
    }
    else if (mode == DAS_STOP)
    {
        /* modes already written above, so counters have already stopped */
    }
    else if (mode == DAS_RESTART)
    {
        #if 0   /* works OK */
        /* configure ADC timer 2; load LSByte then MSB; square wave generator; binary counts */
        outw(0xBE,dev->base_address+DAS_CTR_CTL_REG);
        outw(old_adc_divisor & 0xff,dev->base_address+DAS_CTR2_DATA_REG);          /* write LSB for counter */
        outw(old_adc_divisor >> 8,dev->base_address+DAS_CTR2_DATA_REG);            /* write MSB for counter */
        WAIT_A_WHILE(1); /*?!? wait for regs to settle here*/
        #endif

        /* write divisor only for counters 2 and 1 ?!? assumes ctl regs were written to acheive a stop */
        outw(old_adc_divisor & 0xff,dev->base_address+DAS_CTR2_DATA_REG);          /* write LSB for counter */
        WAIT_A_WHILE(1);                                     /*?!? wait for regs to settle here*/
        outw(old_adc_divisor >> 8,dev->base_address+DAS_CTR2_DATA_REG);            /* write MSB for counter */
        WAIT_A_WHILE(1); /*?!? wait for regs to settle here*/
        outw(old_aux_divisor & 0xff,dev->base_address+DAS_CTR1_DATA_REG);                                 /* write LSB */
        WAIT_A_WHILE(1);                                     /*?!? wait for regs to settle here*/
        outw(old_aux_divisor >> 8,dev->base_address+DAS_CTR1_DATA_REG);                                    /* write MSB */
        WAIT_A_WHILE(1);                                                     /* wait for regs to settle here */
    }
}   /* das16s_program_timers */


/* modify configurable parameters of the card via its control register, etc. */
/* FIXME: phase out arg and store global config on ioctl */
/* note: this calls das16s_program_timers() in 'start' mode which stores adc divisors */
/*******************************************/
int das16s_program_adc()
{
    das16s_dev_t *dev;                                                        /* to access das16s struct */
    unsigned short bitmask;
    /*    int numtries;*/

    DEBUG(3,"das16s_program_adc()\n");

    if (dev_list == NULL)                                                     /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "das16s_program_adc(): module has not been initialized properly!");
        return ERROR;
    }
    else
        dev = (das16s_dev_t *) dev_list->priv;                                /* get our private struct for this card */

    /* stop ints by specifying no int source, soft trig, and reset int enable bit */
    outw(0,dev->base_address+DAS_STATUS_REG);   /* write trig source ?!?*/

    /* stop pacer counters (2 and 1) by writing mode ?!? */
    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_STOP);

    /* inital card config: read the OR'd constants */
    outw(dev->volt_range|dev->dio_lower_dir|dev->dio_upper_dir|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,dev->base_address+DAS_CTRL_REG);

#if 0   /* removed for fix 10 */
    if (dev->channel_scan > 1)                                              /* user wants a scan of several channels */
        outw(ENAB_INT|INT_SRC_END_SCAN|TRIG_SRC_CTR2,dev->base_address+DAS_STATUS_REG); /* enable ints */
    else
        ; /* FIXME: interrupt on FFNE  ?!? */
#endif

    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_START);

    outw(0,dev->base_address+DAS_CONV_REG);          /* write start conversion reg. Clears FIFO overrun & puts a conversion in FIFO?!? */

    /* wait for 'fifo not empty' flag to set (from first conversion ?!?) */
    while (!(inw(dev->base_address+DAS_CTRL_REG) & FIFO_NOT_EMPTY_MASK))
        ;

    /* configure channel scan & write dig out value. mux==0 */
    outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_SCAN_REG); 
    WAIT_A_WHILE(1);                          /* ?!? waste some CPU cycles waiting for regs to settle here ?!? */

    DAS_CLEAR_FIFO(dev->base_address);                                        /* clear the fifo by reading samples out of it */

    bitmask = inw(dev->base_address+DAS_STATUS_REG);    /* ?!? this helps */  /* rewrite status reg */
    outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);   /* clear int by rewriting ctrl reg */
    outw(bitmask,dev->base_address+DAS_STATUS_REG);

/* new w/ fix 10 */
    if (dev->channel_scan > 1)                                              /* user wants a scan of several channels */
        outw(ENAB_INT|INT_SRC_END_SCAN|TRIG_SRC_CTR2,dev->base_address+DAS_STATUS_REG); /* enable ints */
    else
        ; /* FIXME: interrupt on FFNE  ?!? */

#if 0
    /* test of polled ADC conversion on chan 0 - works OK */
    /* stop ints by specifying no int source, soft trig, and reset int enable bit */
    outw(0,dev->base_address+DAS_STATUS_REG);

    /* configure channel scan (this also writes DIO values) */
    outw(SCAN_END_0|SCAN_BEGIN_0,dev->base_address+DAS_SCAN_REG);

    /* inital card config: read the OR'd constants */
    outw(GAIN_BI_10V|DIO_LWR_OUTPUT|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,dev->base_address+DAS_CTRL_REG);

    /* wait for fifo to clear/empty */
    while (inw(dev->base_address+DAS_CTRL_REG) & FIFO_NOT_EMPTY_MASK)
        inw(dev->base_address+DAS_CONV_REG);                            /* gets samplers out of fifo */

    WAIT_A_WHILE(dev->base_address,1000); /* waste 1/2 million CPU cycles waiting for regs to settle here ?!? */

    /* write start conversion reg (clears FIFO overrun) */
    outw(0,dev->base_address+DAS_CONV_REG);

    numtries = 0;
    while (numtries++ < 60000)
        if (inw(dev->base_address+DAS_STATUS_REG) & END_CONVERSION_MASK)
            break;

    if (numtries < 60000)
        printk("%u tries. val = %u\n",numtries,inw(dev->base_address+DAS_CONV_REG));
    else
        printk("too many tries.\n");

    /* inital card config: read the OR'd constants */
    outw(GAIN_BI_10V|DIO_LWR_OUTPUT|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,dev->base_address+DAS_CTRL_REG);

    /* enable ints */
    outw(ENAB_INT|INT_SRC_CTR_2,dev->base_address+DAS_STATUS_REG);

    outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);
    outw(ENAB_INT|INT_SRC_CTR_2,dev->base_address+DAS_STATUS_REG);

#endif

#if 0
    /* test of irq generation by ctr2 - works OK */
    /* stop ints by specifying no int source, soft trig, and reset int enable bit */
    outw(0,dev->base_address+DAS_STATUS_REG);

    /* inital card config: read the OR'd constants */
    outw(GAIN_BI_10V|DIO_LWR_OUTPUT|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,dev->base_address+DAS_CTRL_REG);

    /* enable ints */
    outw(ENAB_INT|INT_SRC_CTR_2,dev->base_address+DAS_STATUS_REG);

    outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);
    outw(ENAB_INT|INT_SRC_CTR_2,dev->base_address+DAS_STATUS_REG);

    /* write modes for counters 2 and 1 */
    outw(COUNTER_2|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN,dev->base_address+DAS_CTR_CTL_REG);
    WAIT_A_WHILE(1);    
    outw(COUNTER_1|CTR_LOAD_LSB_MSB|CTR_SQ_WAVE_GEN,dev->base_address+DAS_CTR_CTL_REG);
    WAIT_A_WHILE(1);

    /* write divisor for counters 2 and 1 - this yields a freq of 1.5 Hz */
    outw(65535/*adc_divisor*/ & 0xff,dev->base_address+DAS_CTR2_DATA_REG);          /* write LSB for counter */
    WAIT_A_WHILE(1);                                     /*?!? wait for regs to settle here*/
    outw(65535/*adc_divisor*/ >> 8,dev->base_address+DAS_CTR2_DATA_REG);            /* write MSB for counter */
    WAIT_A_WHILE(1); /*?!? wait for regs to settle here*/
    outw(100 & 0xff,dev->base_address+DAS_CTR1_DATA_REG);                                 /* write LSB */
    WAIT_A_WHILE(1);                                     /*?!? wait for regs to settle here*/
    outw(0 >> 8,dev->base_address+DAS_CTR1_DATA_REG);                                    /* write MSB */
    WAIT_A_WHILE(1);
#endif

#if 0   /* test of external irqs */
    /* stop ints by specifying no Int source and resetting int enable bit */
    outw(0,dev->base_address+DAS_STATUS_REG);   /* write trig source ?!?*/

    /* inital card config: read the OR'd constants */
    outw(GAIN_BI_10V|DIO_LWR_OUTPUT|ENAB_CTR1_10MHZ|ENAB_CTR0_100_KHZ,dev->base_address+DAS_CTRL_REG);

    /* enable ints */
    outw(ENAB_INT|INT_SRC_EXTERN,dev->base_address+DAS_STATUS_REG);

    outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);
    outw(ENAB_INT|INT_SRC_EXTERN,dev->base_address+DAS_STATUS_REG);

#endif

    return 0;
}   /* das16s_program_adc */


/* This handler is designed for channel scans - it compensates for unfinished channel scans by just counting samples */
/* burst mode here refers to the use of an external multiplexer, addressed w/ DIO bits 0,1 . Currently we can get up to MAX_BANKS*/
/* channel scans per burst */
/* num_mux_banks is total chan scans in a given burst, dev->mux is count of which scan we're currently working on. */
/* Note: there is a problem which only happens at LOW counter frequency, like 30 Hz : */
/* channel scan interrupts come before scan is done! Code to handle this anomaly has been taken out of this handler  ! */
/* Handler does not yet take samples on one individual channel, etc... */
/*******************************************/
unsigned int das16s_irq_handler(unsigned int irq, struct pt_regs *regs)
{
    das16s_dev_t *dev;                                                        /* to access das16s struct */
    int i, ffne;
    /*u16 *sample_bank_p;*/
    static u16 stored_bank[16];     /* for testing ?!?*/
    static int firstime=1;     /* for testing ?!?*/
    /*static unsigned char in_irq_handler=0;*/                                             /* reentrance flag */

    if (in_fop_ioctl)                         /* happens if a burst START comes before previous burst is finished */
        if (testing)                          /* We don't return here because 2nd START is ignored in ioctl(), thus this data is valid */
            DEBUG(0,"das16s Error: in irq handler during fop_ioctl()!\n");

    if (in_irq_handler)
    {
        DEBUG(0,"Re-entered das16s_irq_handler, Aborting!\n");
        return 1;
    }
    else
        in_irq_handler = TRUE;                                                /* set flag to prevent reentrance into this handler */

    if (dev_list == NULL)                                                     /* if there is a list of instances of the card */
    {
        printk(KERN_WARNING "das16s_irq_handler(): module has not been initialized properly!");
        return ERROR;
    }
    else
        dev = (das16s_dev_t *) dev_list->priv;                              /* get our private struct for this card */

    if (!(EXT_INT_STATUS_MASK & inw(dev->base_address+DAS_STATUS_REG)))       /* make sure this wasn't external irq */
    {
      if (dev->channel_scan != 1)                                           /* user wants one or more banks (scans) of samples */
      {
        DEBUG(3,"\nINT ");

        /* must write the control register to enable further interrupts (clears INTB in status reg) */
        outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG); /* fencepost */
        /* STI here ?!? */

        /*printk("STAT:%d ",inw(dev->base_address+DAS_STATUS_REG));*/

        while ((ffne = inw(dev->base_address+DAS_CTRL_REG) & FIFO_NOT_EMPTY_MASK))    /* read all available samples in fifo */
        {
            /* must write the control register here (clears INTB in status reg). without this we must wait twice as many irqs */
            outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);
            if (dev->chan_cnt < dev->channel_scan)                            /* end of scan, don't overflow buffer */
            {
                /* get a single sample out of FIFO and put it in current packet in data table */
                INSERT_ADC_RESULT(dev,inw(dev->base_address+DAS_CONV_REG),dev->chan_cnt++);
                DEBUG(3,"%d_f%d ",dev->chan_cnt-1,ffne);
                DEBUG(3,"%d ",dev->result_table[dev->result_table_tail][ADC_DATA_OFFSET+(dev->chan_cnt-1)]);
                /*printk("%d_f%d ",dev->chan_cnt-1,FIFO_NOT_EMPTY_MASK & inw(dev->base_address+DAS_CTRL_REG));*/
                if (FIFO_OVERFLOW_MASK & inw(dev->base_address+DAS_STATUS_REG)) /* new */
                    printk(KERN_DEBUG "das16s: FIFO overrun!\n");
            }
            else
            {
                /*DEBUG(3,"Too many fifo reads!\n");*/       /* just for safety */
                /* warning: don't loop here - we went infinite once ! */
                DEBUG(3,"+ ");
                break;                                                  /* leave extra samples in fifo until later */
            }
        }
        DEBUG(3,"endw\n");

        if (testing)
        {
            if (firstime)   /* for testing ?!?*/
            {
                    memcpy(stored_bank,&(dev->result_table[dev->result_table_tail][ADC_DATA_OFFSET+0]),32);
                    firstime=0;
            }
            else    /* for testing ?!?*/
            {   /* check every bank for first sample > +/-100 from first ever sample */
                if (abs(dev->result_table[dev->result_table_tail][ADC_DATA_OFFSET] - stored_bank[0]) > 100)
                {
                    printk("sample glitch! (bank %d)\norig: ",dev->mux);
                    for(i=0; i < 16; i++)
                        printk("%05d ",stored_bank[i]);
                    printk("\ncurr: ");
                    for(i=0; i < 16; i++)
                        printk("%05d ",dev->result_table[dev->result_table_tail][ADC_DATA_OFFSET+i]);
                    printk("\n");                    
                    if (in_fop_ioctl)                                           
                        printk("glitch was during fop_ioctl()!\n"); /* we preempted an ioctl - user app reconfig before end of burst */
                }
            }
        }

        if (dev->chan_cnt == dev->channel_scan)                             /* got a bank of samples. store and reset  */
        {
            SET_RESULT_MUX(dev,dev->mux);                                   /* stamp result packet w/ mux value */
            /*sample_bank_p = GET_CURRENT_RESULT_PACKET(dev); for printk's below*/
            /*DEBUG(3,"mux=%d %d samples: ",sample_bank_p[HDR_MUX_OFFSET],dev->chan_cnt);*/
            /*for (i=0; i < dev->chan_cnt; i++)                       
                printk("%u ",sample_bank_p[ADC_DATA_OFFSET+i]); */    /* just for testing ?!? */
            DEBUG(3,"\n");                                           /* just for testing ?!? */
            dev->digital_input_word |= ((inw(dev->base_address+DAS_DIO_REG) >> 8)&0xF) << (4*dev->mux); /* read dig in's from each bank*/
            if (dev->mux == dev->num_mux_banks-1)                           /* done w/ a burst */
            {
                dev->mux = 0;                                               /* reset to first bank of external mux */
                if (dev->burst_acq)                                         /* if in burst mode we stop after getting all banks */
                {
                    /* was 6th fix, commented out in 7th fix, back in for 9th */
                    /* stop ints by specifying no int source, soft trig, and reset int enable bit *//* new w/ 6th glitch workaround */
                    outw(0,dev->base_address+DAS_STATUS_REG);               /* new w/ 6th glitch workaround */

                    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_STOP); /* stop pacer counters */
                    /* new stuff below */
                    dev->dio_output_val &= (~0x03 << 12);                /* clear 2 LSBs of upper nibble of MSByte, will be ORed w/ mux */
                    dev->dio_output_val |= (dev->mux << 12);                 /* combine dig out word w/ mux val */
                    outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_DIO_REG); /* write digout, reset scan */
                    burst_in_process = FALSE;                                 /* reset global flag - done w/ burst job */
                }
                else                                                        /* in continuous mode, write external mux */
                {
                    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_STOP); /* stop pacer counters */
                    dev->dio_output_val &= (~0x03 << 12);              /* clear 2 LSBs of upper nibble of MSByte, will be ORed w/ mux */
                    dev->dio_output_val |= (dev->mux << 12);                 /* combine dig out word w/ mux val */
                    outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_DIO_REG); /* write digout, reset scan */
                    DAS_CLEAR_FIFO(dev->base_address);                      /* must clear the fifo or we mess up the sample count */
                    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_START); /* stop pacer counters */
                }
                SET_DIG_IN_RESULT(dev,dev->digital_input_word);         /* put (packed) values into current packet of result table */
                dev->digital_input_word = 0;                            /* reset for next burst */
            }
            else                                                        /* not done w/ burst */
            {
                if (dev->multiplexing)                                  /* must switch external MUX */
                {
                    /* FIXME: read dig in's from each bank in turn */
                    dev->mux += 1;                                          /* increment external mux */
                    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_STOP); /* stop pacer counters */
                    dev->dio_output_val &= (~0x03 << 12);                /* clear 2 LSBs of MSByte, value will be ORed w/ mux */
                    dev->dio_output_val |= (dev->mux << 12);                 /* combine dig out word w/ mux val */
                    /* write dig out value & reset the channel scan */
                    outw(dev->dio_output_val|dev->scan_end|dev->scan_begin,dev->base_address+DAS_DIO_REG); 
                    DAS_CLEAR_FIFO(dev->base_address);                      /* must clear the fifo or we mess up the sample count */
                    das16s_program_timers(dev->adc_rate_divisor,dev->aux_adc_rate_divisor, DAS_START); /* start pacer counters */
                    DEBUG(3,"muxswitch:%d\n",dev->mux);
                }
            }
            NEXT_ADC_RESULT_PACKET(dev);                                /* step to next result packet */
            dev->chan_cnt = 0;                                              /* reset count. start new scan */
        }   /* endif got a bank */
      }
      else                                                                  /* single chan conversion mode */
        {
            /* handle contin/burst modes for single chan conversions */
        }
    }   /* end of our irq */
    else
    {
        DEBUG(3,"EXT INT\n");
        /* must write the control register to enable further interrupts (clears INTB in status reg) */
        outw(inw(dev->base_address+DAS_CTRL_REG),dev->base_address+DAS_CTRL_REG);
    }

    rtl_hard_enable_irq(dev->irq);                                            /* reenable this irq @ mask reg for further handler calls*/
    in_irq_handler = FALSE;                                                   /* reset flag, we're no longer in this handler */

    return 0;                                                               /* normal ret val */
}   /* das16s_irq_handler */
