/*======================================================================

    das16s.h    v0.03    Apr 26 2000

    PCM-DAS16s/16 real-time PCMCIA client driver

    for Real Time Linux kernels only, at present.
    see http://www.rtlinux.org

    by Steve Rosenbluth (stever8@earthlink.net, stever@la.creatureshop.henson.com)

    Copyright (C) 2000 Jim Henson's Creature Shop (Los Angeles)

    The contents of this file are subject to the Mozilla Public
    License Version 1.1 (the "License"); you may not use this file
    except in compliance with the License. You may obtain a copy of
    the License at http://www.mozilla.org/MPL/

    Software distributed under the License is distributed on an "AS
    IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
    implied. See the License for the specific language governing
    rights and limitations under the License.

    The developer of the initial PCMCIA code is David A. Hinds
    <dhinds@pcmcia.sourceforge.org>.  Portions created by David A. Hinds
    are Copyright (C) 1999 David A. Hinds.  All Rights Reserved.    

    Thanks to Computer Boards, Inc. for providing register documentation
======================================================================*/
/*  epson dot-matrix printer condense code (137 chars per row) */

#include "das16s-config.h"

/* DAS16s register offsets */
#define DAS_CONV_REG            0x00
#define DAS_SCAN_REG            0x02
#define DAS_DIO_REG             0x02
#define DAS_STATUS_REG          0x04
#define DAS_CTRL_REG            0x06
#define DAS_CTR0_DATA_REG       0x08
#define DAS_CTR1_DATA_REG       0x0A
#define DAS_CTR2_DATA_REG       0x0C
#define DAS_CTR_CTL_REG         0x0E
/* OR masks for setting bits in control register (base+6) */
#define ENAB_EXTERN_TRIG        0x01
#define ENAB_CTR0_100_KHZ       0x02
#define ENAB_CTR1_1MHZ          0x00
#define ENAB_CTR1_10MHZ         0x04
#define ENAB_PWR_DWN            0x20
#define DIO_LWR_OUTPUT          0x40
#define DIO_UPR_OUTPUT          0x80
#define GAIN_BI_10V             0x800
#define GAIN_BI_5V              0x000
#define GAIN_BI_2_5V            0x100
#define GAIN_BI_1_25V           0x200
#define GAIN_UNI_10V            0x400
#define GAIN_UNI_5V             0x500
#define GAIN_UNI_2_5V           0x600
#define GAIN_UNI_1_25V          0x700
#define ENAB_OFFSET_CALIB       0x1000
#define ENAB_GAIN_CALIB         0x2000
/* AND masks for checking bits in control register (base+6) */
#define FIFO_NOT_EMPTY_MASK     0x08
#define FIFO_OVERRUN_MASK       0x400
/* OR masks for setting bits in status reg */
#define TRIG_SRC_CTR2           0X300
#define INT_SRC_END_SCAN        0x5000
#define INT_SRC_CTR_2           0x1000
#define INT_SRC_EXTERN          0x2000
#define ENAB_INT                0x8000
#define INTB_CLEAR              0x10
/* OR masks for setting bits in DIO/SCAN reg */
#define SCAN_END_15             0xf0
#define SCAN_END_7              0x70
#define SCAN_END_0              0x00
#define SCAN_BEGIN_0            0x00
/* AND masks for checking bits in status reg */
#define EXT_INT_STATUS_MASK     0x10    /* INTB */
#define END_CONVERSION_MASK     0x80
#define FIFO_OVERFLOW_MASK      0x400
/* OR masks for setting bits in counter config register */
#define COUNTER_0               0x00
#define COUNTER_1               0x40
#define COUNTER_2               0x80
#define CTR_LOAD_LSB_MSB        0x30
#define CTR_SQ_WAVE_GEN         0x0E

#define DAS16S_IO_SPACE         16                                          /* io footprint */
#define MAX_FIFO_SAMPLES        512                                         /* card holds this many samples in its hardware fifo */
#define ERROR                   -1
#define DIO_OUT                 1
#define DIO_IN                  0


/*
   A dev_link_t structure has fields for most things that are needed
   to keep track of a socket, but there will usually be some device
   specific information that also needs to be kept track of.  The
   'priv' pointer in a dev_link_t structure can be used to point to
   a device-specific private data structure, like this.

   To simplify the data structure handling, we actually include the
   dev_link_t structure in the device's private data structure.

   A driver needs to provide a dev_node_t structure for each device
   on a card.  In some cases, there is only one device per card (for
   example, ethernet cards, modems).  In other cases, there may be
   many actual or logical devices (SCSI adapters, memory cards with
   multiple partitions).  The dev_node_t structures need to be kept
   in a linked list starting at the 'dev' field of a dev_link_t
   structure.  We allocate them in the card's private data structure,
   because they generally shouldn't be allocated dynamically.

   In this case, we also provide a flag to indicate if a device is
   "stopped" due to a power management event, or card ejection.  The
   device IO routines can use a flag like this to throttle IO to a
   card that is not ready to accept it.

   The bus_operations pointer is used on platforms for which we need
   to use special socket-specific versions of normal IO primitives
   (inb, outb, readb, writeb, etc) for card IO.
*/

typedef unsigned char bool;

typedef struct das16s_dev_t {
    dev_link_t		link;
    dev_node_t		node;
    int			stop;
    struct bus_operations *bus;

    ioaddr_t base_address;
    int irq;
    u16 adc_rate_divisor;
    u16 aux_adc_rate_divisor;
    int channel_scan;                                                         /* num channels to sample. valid 1-16 */
    u16 volt_range;                                                           /* OR bitmask for voltage range (gain) */
    bool start;
    int num_mux_banks;  /* total num banks (chan scans) of adc chans through which driver should multiplex */
    bool burst_acq;  /* if true: do a burst of all samples and stop sampling, else continuous acq */
    int num_dig_in_banks;                                                   /* how many banks of digital input bits: 1 to 4 nibbles */
    u16 dio_output_val;  /* val to be written to the digital output port. 2 LSBs reserved if external MUX in use */
    char multiplexing;                                                        /* whether using digital out bits 0,1 for external MUX */
    char send_dig_ins;                                                        /* flag for irq handler that inputs have changed */
    int chan_cnt;                                    /* count of individual adc channel conversions */
    int mux;                                         /* current bank of channel scans on which we get conversions. external mux addr */
    /*unsigned int sample_bank[MAX_CHANNELS+SAMPLE_HEADER_SIZE];*/       /* conversion value storage, init first element to magic ID */
    bool error_correcting;
    u16 dio_lower_dir;                                           /* whether bits go out or come in */
    u16 dio_upper_dir;
    u16 scan_begin;                                              /* limits of channel scan */
    u16 scan_end;
    struct wait_queue read_q;
    u16 result_table[MAX_SAMPLE_PACKETS+1][RESULT_PACK_SIZE_SI]; /* table of result packets */
    int result_table_head;                                                  /* index to next available packet in result table */
    int result_table_tail;                                                  /* index to next free slot in result table */
    int result_pack_avail;
    u16 digital_input_word;                                                 /* temp storage for digital input */
/*    int available_in_fifo; */                                                 /* count of available samples in result table */
} das16s_dev_t;
